﻿Type=StaticCode
Version=1.5
ModulesStructureVersion=1
B4R=true
@EndOfDesignText@


Sub Process_Globals
	Private controlSocket As WiFiSocket
	Private dataSocket As WiFiSocket
	Private AStreamControl, astreamData As AsyncStreams
	Private EOL() As Byte = Array As Byte(13, 10)
	Private const AFTER_CONNECTION = 0, AFTER_USER = 1, AFTER_PASS = 2, BEFORE_RETR = 3, AFTER_RETR = 4 As Byte
	Private const IDLE_LOGGEDIN = 5, BEFORE_STOR = 6, AFTER_STOR = 7 As Byte
	Private state As Byte
	Private bc As ByteConverter
	Private FileBuffer(1000) As Byte
	Private FileBufferIndex As UInt
End Sub

Public Sub ConnectToServer (IP() As Byte, Port As UInt, User() As Byte, Password() As Byte)
	If controlSocket.ConnectIP(IP, Port) = False Then
		Main.FTP_LogIn(False)
		Return
	End If
	AStreamControl.Initialize(controlSocket.Stream, "astreamControl_NewData", "astreamControl_Error")
	GlobalStore.Put(0, User)
	GlobalStore.Put(1, Password)
	GlobalStore.Put(2, IP) 'should stay there all the time
	
	state = AFTER_CONNECTION
End Sub

Public Sub Download(Path() As Byte)
	If CheckIdle Then Return
	GlobalStore.Put(0, Path)
	AStreamControl.Write("PASV").Write(EOL)
	state = BEFORE_RETR
End Sub

Public Sub Upload(Path() As Byte, Data() As Byte)
	If CheckIdle Then Return
	bc.ArrayCopy(Data, FileBuffer)
	FileBufferIndex = Data.Length
	state = BEFORE_STOR
	GlobalStore.Put(0, Path)
	AStreamControl.Write("PASV").Write(EOL)
End Sub

Private Sub CheckIdle As Boolean
	If state <> IDLE_LOGGEDIN Then
		Log("Wait for previous task to complete.")
		Return True
	End If
	Return False
End Sub

Public Sub Close
	If controlSocket.Connected Then AStreamControl.Write("QUIT ").Write(EOL)
	controlSocket.Close
End Sub



Private Sub AStreamControl_NewData (Buffer() As Byte)
	'we need to take care of incomplete messages.
	'the last non-complete line is stored in slot 4 and is then appended to the current message.
	If GlobalStore.Slot4.Length > 0 Then
		'append the new message to the existing leftover.
		Buffer = JoinBytes(Array(GlobalStore.Slot4, Buffer))
		GlobalStore.Put(4, Array As Byte())
	End If
	If bc.EndsWith(Buffer, EOL) = False Then
		'message is not complete.
		'remove the last part
		Dim i As Int = bc.LastIndexOf(Buffer, EOL)
		If i > -1 Then
			GlobalStore.Put(4, bc.SubString2(Buffer, i + EOL.Length, Buffer.Length))
			Buffer = bc.SubString2(Buffer, 0, i)
		Else
			'The message doesn't include any complete sentence. Store it and return
			GlobalStore.Put(4, Buffer)
			Return 
		End If
	End If
	
	For Each line() As Byte In bc.Split(Buffer, EOL)
		If line.Length < 3 Then Continue
		Dim s As String = bc.StringFromBytes(bc.SubString2(line, 0, 3))
		If IsNumber(s) Then
			Dim st As Int = s
			If st >= 100 And st <= 999 Then
				HandleServerResponse(st, Buffer)
			End If
		End If
	Next
End Sub

Private Sub HandleServerResponse(status As Int, Buffer() As Byte)
	If status = 227 Then
		OpenDataConnection(Buffer)
	End If
	Select state
		Case AFTER_CONNECTION
			AStreamControl.Write("USER ").Write(GlobalStore.Slot0).Write(EOL)
			state = AFTER_USER
		Case AFTER_USER
			If status = 331 Then
				AStreamControl.Write("PASS ").Write(GlobalStore.Slot1).Write(EOL)
				state = AFTER_PASS
			Else if status = 220 Then
				'do nothing
			Else
				Main.FTP_LogIn(False)
			End If
		Case AFTER_PASS
			If status = 230 Then
				state = IDLE_LOGGEDIN
				Main.FTP_LogIn(True)
			Else
				Main.FTP_LogIn(False)
			End If
		Case BEFORE_RETR
			FileBufferIndex = 0
			AStreamControl.Write("RETR ").Write(GlobalStore.Slot0).Write(EOL)
			state = AFTER_RETR
		Case AFTER_RETR
			If status <> 226 And status <> 150 Then
				state = IDLE_LOGGEDIN
				'error
				Main.FTP_DownloadComplete(False, GlobalStore.Slot0, Null)
			End If
		Case BEFORE_STOR
			AStreamControl.Write("STOR ").Write(GlobalStore.Slot0).Write(EOL)
			state = AFTER_STOR
		Case AFTER_STOR
			If status = 150 Then
				'send the data
				astreamData.Write2(FileBuffer, 0, FileBufferIndex)
				dataSocket.Stream.Flush
				dataSocket.Close
			Else If status = 226 Then
				state = IDLE_LOGGEDIN
				Main.FTP_UploadComplete(True, GlobalStore.Slot0)
			Else
				state = IDLE_LOGGEDIN
				Main.FTP_UploadComplete(False, GlobalStore.Slot0)
			End If
	End Select
End Sub

Private Sub OpenDataConnection(Buffer() As Byte)
	If dataSocket.Connected Then dataSocket.Close
	'find the port number
	Dim i As Int = 0
	Dim port As UInt
	For Each s() As Byte In bc.Split(Buffer, ",")
		If i = 4 Then
			port = 256 * bc.StringFromBytes(s)
		Else if i = 5 Then
			port = port + bc.StringFromBytes(s)
		End If
		i = i + 1
	Next
	If dataSocket.ConnectIP(GlobalStore.Slot2, port) Then
		astreamData.Initialize(dataSocket.Stream, "astreamData_NewData", "astreamData_Error")
	Else
		Log("Data socket failed to connect")
	End If
End Sub

Private Sub astreamData_NewData (Buffer() As Byte)
	If FileBufferIndex + Buffer.Length > FileBuffer.Length Then
		Log("FileBuffer too small!")
		dataSocket.Close
		Return
	End If
	'copy the data to FileBuffer
	bc.ArrayCopy2(Buffer, 0, FileBuffer, FileBufferIndex, Buffer.Length)
	FileBufferIndex = FileBufferIndex + Buffer.Length
End Sub

Private Sub astreamData_Error
	If state = AFTER_RETR Then
		state = IDLE_LOGGEDIN
		Main.FTP_DownloadComplete(True, GlobalStore.Slot0, bc.SubString2(FileBuffer, 0, FileBufferIndex))
	End If
End Sub

Private Sub AStreamControl_Error
	Log("control closed")
End Sub